/*******************************************************************************
 *
 * ============================================================================
 *
 * Copyright (c) 2001-2016 Primeton Technologies, Ltd. All rights reserved.
 *
 * Created on 2017年2月16日 下午3:30:05
 *******************************************************************************/

package com.primeton.devops.uc.service;

import java.util.Date;
import java.util.List;
import java.util.concurrent.Callable;

import org.apache.commons.lang.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.primeton.devops.common.ConnectionContextThreadLocal;
import com.primeton.devops.common.DataContextThreadLocal;
import com.primeton.devops.common.dao.TenantBaseDao;
import com.primeton.devops.common.logger.IOperationLogger;
import com.primeton.devops.common.logger.OperationLoggerFactory;
import com.primeton.devops.common.service.JwtTokenService;
import com.primeton.devops.common.util.CommonUtil;
import com.primeton.devops.common.util.PasswordCryptoUtil;
import com.primeton.devops.common.util.ValidateUtil;
import com.primeton.devops.integration.jira.JiraProjectManageToolService;
import com.primeton.devops.integration.ldap.LdapController;
import com.primeton.devops.integration.ldap.LdapServerFactory;
import com.primeton.devops.specs.api.IDistributedLockService;
import com.primeton.devops.specs.api.pcm.ILdapServer;
import com.primeton.devops.specs.api.pcm.service.IBusinessDictService;
import com.primeton.devops.specs.api.pcm.service.ISysConfService;
import com.primeton.devops.specs.api.uc.service.IUserService;
import com.primeton.devops.specs.constant.Constants;
import com.primeton.devops.specs.constant.pcm.RunTargetType;
import com.primeton.devops.specs.constant.pcm.ScopeType;
import com.primeton.devops.specs.constant.pcm.SystemType;
import com.primeton.devops.specs.constant.uc.EmployeeStatus;
import com.primeton.devops.specs.constant.uc.UcConstants;
import com.primeton.devops.specs.exception.DevOps;
import com.primeton.devops.specs.exception.DevOpsException;
import com.primeton.devops.specs.model.UserInfo;
import com.primeton.devops.specs.model.pcm.Infra3rdSystem;
import com.primeton.devops.specs.model.pcm.TargetRunInfo;
import com.primeton.devops.specs.model.uc.Employee;
import com.primeton.devops.specs.model.uc.User;
import com.primeton.devops.specs.model.uc.UserResult;

import io.jsonwebtoken.Claims;
import io.jsonwebtoken.impl.DefaultClaims;

/**
 * UserService.
 *
 * @author guohuaicheng (mailto:guohc@primeton.com)
 */
@Service
public class UserService implements IUserService {

	private static IOperationLogger operationLogger = OperationLoggerFactory.getLogger(UserService.class);
	
	private static final String CONSECUTIVE_LOGIN_FAIL_COUNT_KEY = "consecutiveLoginFailCount";
	
	private static final String LOGIN_FAIL_TIME_KEY = "loginFailTime";
	
	private static final String USER_LOCKED_FLAG_KEY = "userLockedFlag";

	@Autowired
	private IBusinessDictService businessDictService;

	@Autowired
	private EmployeeService employeeService;
	
	@Autowired
	private TenantBaseDao baseDao;

	@Autowired
	private ISysConfService sysConfService;

	@Autowired
	private JwtTokenService jwtTokenService;
	
	@Autowired
	private JiraProjectManageToolService jiraProjectManageToolService;	
	
	@Autowired
	private LdapController ldapController;

	@Autowired
	private IDistributedLockService lockService;

	@Override
	public UserResult login(String userName, String password) throws Exception {
		ValidateUtil.assertNotNullOrEmpty(userName, "userName");
		ValidateUtil.assertNotNullOrEmpty(password, "password");
		TargetRunInfo targetRunInfo = TargetRunInfo.FACTORY.create();
		targetRunInfo.setTargetType(RunTargetType.USER.getDbValue());
		targetRunInfo.setTargetId(userName);
		targetRunInfo = baseDao.getEntity(targetRunInfo);
		//判断是否锁定
		if (targetRunInfo != null && "true".equals(CommonUtil.getExtendValue(targetRunInfo, USER_LOCKED_FLAG_KEY))) {
			int maxConsecutiveLoginFailCount = getMaxConsecutiveLoginFailCount();
			if (maxConsecutiveLoginFailCount <= 0) {
				unlockLogin(userName);
			}
			int loginFailExceedMaxCountLockTime = getLoginFailExceedMaxCountLockTime();
			if (loginFailExceedMaxCountLockTime <= 0) {
				unlockLogin(userName);
			}
			Date loginFailTime = CommonUtil.formatDate(CommonUtil.getExtendValue(targetRunInfo, LOGIN_FAIL_TIME_KEY));
			int lockTime = 0;
			if (loginFailTime != null) {
				lockTime = (int)(System.currentTimeMillis() - loginFailTime.getTime()) / 1000;
			}
			if (lockTime < 0) {
				lockTime = 0;
			}
			int remainTime = loginFailExceedMaxCountLockTime * 60 - lockTime;
			if (remainTime > 0) {
				throw new DevOpsException(DevOps.UC.USER_LOCKED_BECAUSE_LOGIN_FAIL_EXCEED_MAX_COUNT, 
						userName, loginFailExceedMaxCountLockTime, maxConsecutiveLoginFailCount, remainTime);
			} else {
				unlockLogin(userName);
			}
		}
		User user = authenticate(userName, password);
		doUnlockLogin(targetRunInfo);
		
		String token = jwtTokenService.generateToken(user.getEmployee().getEmpId(), userName, user.getEmployee().getEmpName(), null);
		UserResult result = new UserResult();
		result.setEmpId(user.getEmployee().getEmpId());
		result.setUserName(userName);
		result.setToken(token);
		DataContextThreadLocal.setUserInfo(user.getEmployee().getEmpId(), userName, user.getEmployee().getEmpName(), null);
		DataContextThreadLocal.setToken(token);
		result.setResultCode(0);
		operationLogger.logOther(ScopeType.PLATFORM.getDbValue(), null, "Login: {0}", new Object[] {userName}, 
				User.QNAME, "userName", userName);
		return result;
	}
	
	//登录验证
	protected User authenticate(String userName, String password) throws Exception {
		if (userName.equals(UcConstants.USER_NAME_SYSADMIN)) {//系统管理员
			return doAuthenticate(userName, password);
		}
		if (!"true".equals(businessDictService.getConfigParamValue(null, "Uc.AutoCreateUserWhenLogin"))) {
			if (baseDao.getEntity(User.QNAME, userName) == null) {// 用户不存在
				throw new DevOpsException(DevOps.UC.USER_NOT_EXISTED, userName);
			}
		}
		String loginAuthenticateModel = businessDictService.getConfigParamValue(null, "Uc.LoginAuthenticateModel");
		ILdapServer ldapServer = null;
		boolean getLdapServer = false;
		if (StringUtils.isBlank(loginAuthenticateModel)) {
			ldapServer = getLdapserver(userName);
			getLdapServer = true;
			if (ldapServer == null) {
				return doAuthenticate(userName, password);
			} else {
				return doAuthenticateWithLdap(ldapServer, userName, password);
			}
		}
		if ("local".equals(loginAuthenticateModel)) {
			return doAuthenticate(userName, password);
		}
		if ("jira".equals(loginAuthenticateModel)) {
			return doAuthenticateWithJira(userName, password);
		}
		if (!getLdapServer) {
			ldapServer = getLdapserver(userName);
		}
		if (ldapServer == null) {
			throw new DevOpsException(DevOps.PCM.SYSCONF_NOT_CONFIG, "ldap");
		}
		return doAuthenticateWithLdap(ldapServer, userName, password);
	}
	
	protected User doAuthenticate(String userName, String password) throws Exception {
		User user = baseDao.getEntity(User.QNAME, userName);
		if (user == null) {// 用户不存在
			throw new DevOpsException(DevOps.UC.USER_NOT_EXISTED, userName);
		} else {
			if (user.getEmployee() == null 
					|| Constants.DB_LOGIC_DEL_FLAG.equals(user.getDelFlag())
					|| Constants.DB_LOGIC_DEL_FLAG.equals(user.getEmployee().getDelFlag())
					|| EmployeeStatus.LEAVE.getDbValue().equals(user.getEmployee().getStatus())) {
				throw new DevOpsException(DevOps.UC.ILLEGAL_USER, userName);
			}
			String encodedPassword = PasswordCryptoUtil.encrypt(password);
			if (!encodedPassword.equals(user.getPassword())) {// 密码不正确
				handlePasswordIncorrect(userName);
				throw new DevOpsException(DevOps.UC.USER_PASSWORD_INCORRECT);
			}
		}
		return user;
	}
	
	private int getMaxConsecutiveLoginFailCount() throws Exception {
		return businessDictService.getConfigParamValue(null, "Uc.MaxConsecutiveLoginFailCount", -1, int.class);
	}
	
	private int getLoginFailExceedMaxCountLockTime() throws Exception {
		return businessDictService.getConfigParamValue(null, "Uc.LoginFailExceedMaxCountLockTime", -1, int.class);
	}
	
	protected void handlePasswordIncorrect(String userName) throws Exception {
		int maxConsecutiveLoginFailCount = getMaxConsecutiveLoginFailCount();
		if (maxConsecutiveLoginFailCount <= 0) {
			return;
		}
		int loginFailExceedMaxCountLockTime = getLoginFailExceedMaxCountLockTime();
		if (loginFailExceedMaxCountLockTime <= 0) {
			return;
		}
		lockService.lockAndAutoUnlock("User.login.fail", userName, -1, new Callable<Void>() {
			@Override
			public Void call() throws Exception {
				TargetRunInfo targetRunInfo = TargetRunInfo.FACTORY.create();
				targetRunInfo.setTargetType(RunTargetType.USER.getDbValue());
				targetRunInfo.setTargetId(userName);
				int remainCount = maxConsecutiveLoginFailCount;
				if (baseDao.getEntity(targetRunInfo) == null) {
					remainCount = maxConsecutiveLoginFailCount - 1;
					CommonUtil.setExtendValue(targetRunInfo, CONSECUTIVE_LOGIN_FAIL_COUNT_KEY, 1);
					CommonUtil.setExtendValue(targetRunInfo, LOGIN_FAIL_TIME_KEY, CommonUtil.formatDate(new Date()));
					baseDao.insertEntity(targetRunInfo);
				} else {
					int consecutiveLoginFailCount = CommonUtil.getExtendValue(targetRunInfo, CONSECUTIVE_LOGIN_FAIL_COUNT_KEY, 0);
					consecutiveLoginFailCount++;
					Date loginFailTime = CommonUtil.formatDate(CommonUtil.getExtendValue(targetRunInfo, LOGIN_FAIL_TIME_KEY));
					int alreadyFailedTime = 0;
					if (loginFailTime != null) {
						alreadyFailedTime = (int)(System.currentTimeMillis() - loginFailTime.getTime()) / 1000;
					}
					if (alreadyFailedTime < 0) {
						alreadyFailedTime = 0;
					}
					if (alreadyFailedTime >= loginFailExceedMaxCountLockTime * 60) {
						consecutiveLoginFailCount = 1;
					}
					remainCount = maxConsecutiveLoginFailCount - consecutiveLoginFailCount;
					CommonUtil.setExtendValue(targetRunInfo, CONSECUTIVE_LOGIN_FAIL_COUNT_KEY, consecutiveLoginFailCount);
					CommonUtil.setExtendValue(targetRunInfo, LOGIN_FAIL_TIME_KEY, CommonUtil.formatDate(new Date()));
					if (remainCount <= 0) {
						CommonUtil.setExtendValue(targetRunInfo, USER_LOCKED_FLAG_KEY, "true");
					} else {
						CommonUtil.setExtendValue(targetRunInfo, USER_LOCKED_FLAG_KEY, "false");
					}
					baseDao.updateEntity(targetRunInfo);
				}
				ConnectionContextThreadLocal.commitAndClose();
				if (remainCount == 0) {
					throw new DevOpsException(DevOps.UC.USER_LOCKED_WHEN_LOGIN_FAIL_REACHED_MAX_COUNT, 
							userName, loginFailExceedMaxCountLockTime, maxConsecutiveLoginFailCount);
				} else {
					throw new DevOpsException(DevOps.UC.USER_LOGIN_FAIL_CONTINUE_ATTEMPT, userName, remainCount);
				}
			}
		});
	}
	
	protected User doAuthenticateWithJira(String userName, String password) throws Exception {
		Infra3rdSystem jiraServerConfig = jiraProjectManageToolService.getJiraConfig();
		if (jiraServerConfig == null) {
			throw new DevOpsException(DevOps.PCM.SYSCONF_NOT_CONFIG, "jira");
		}
		if (jiraProjectManageToolService.login(jiraServerConfig, userName, password) != null) {
			User user = baseDao.getEntity(User.QNAME, userName);
			if (user == null) {
				user = jiraProjectManageToolService.get3rdUser(userName);
				employeeService.createEmployee(user.getEmployee());
				user = baseDao.getEntity(User.QNAME, userName);
			} else {
				if (user.getEmployee() == null 
						|| Constants.DB_LOGIC_DEL_FLAG.equals(user.getDelFlag())
						|| Constants.DB_LOGIC_DEL_FLAG.equals(user.getEmployee().getDelFlag())
						|| EmployeeStatus.LEAVE.getDbValue().equals(user.getEmployee().getStatus())) {
					throw new DevOpsException(DevOps.UC.ILLEGAL_USER, userName);
				}
			}
			String encodedPassword = PasswordCryptoUtil.encrypt(password);
			if (!encodedPassword.equals(user.getPassword())) {
				user.setPassword(encodedPassword);
				baseDao.updateEntity(user);
			}
			return user;
		} else {
			if (jiraProjectManageToolService.get3rdUser(userName) == null) {
				throw new DevOpsException(DevOps.UC.USER_NOT_EXISTED, userName);
			} else {
				handlePasswordIncorrect(userName);
				throw new DevOpsException(DevOps.UC.USER_PASSWORD_INCORRECT);
			}
		}
	}

	private ILdapServer getLdapserver(String userName) throws Exception {
		List<Infra3rdSystem> ldapServerList = sysConfService.querySysConfs(SystemType.LDAP_SERVER.getDbValue(), null, null, null);
		if (ldapServerList == null || ldapServerList.isEmpty()) {
			return null;
		}
		for (Infra3rdSystem system : ldapServerList) {
			ILdapServer ldapServer = LdapServerFactory.getLdapServer(system);
			if (ldapServer.getUser(userName) != null) {
				return ldapServer;
			}
		}
		return null;
	}
	
	protected User doAuthenticateWithLdap(ILdapServer ldapServer, String userName, String password) throws Exception {
		if (ldapServer.authenticate(userName, password)) {
			User user = baseDao.getEntity(User.QNAME, userName);
			if (user == null) {
				Employee ldapEmployee = ldapServer.getUser(userName);
				ldapEmployee.setOrganization(ldapController.createOrUpdateOrganization(ldapEmployee.getOrganization()));
				employeeService.createEmployee(ldapEmployee);
				user = baseDao.getEntity(User.QNAME, userName);
			} else {
				if (user.getEmployee() == null 
						|| Constants.DB_LOGIC_DEL_FLAG.equals(user.getDelFlag())
						|| Constants.DB_LOGIC_DEL_FLAG.equals(user.getEmployee().getDelFlag())
						|| EmployeeStatus.LEAVE.getDbValue().equals(user.getEmployee().getStatus())) {
					throw new DevOpsException(DevOps.UC.ILLEGAL_USER, userName);
				}
			}
			String encodedPassword = PasswordCryptoUtil.encrypt(password);
			if (!encodedPassword.equals(user.getPassword())) {
				user.setPassword(encodedPassword);
				baseDao.updateEntity(user);
			}
			return user;
		} else {
			if (ldapServer.getUser(userName) == null) {
				throw new DevOpsException(DevOps.UC.USER_NOT_EXISTED, userName);
			} else {
				handlePasswordIncorrect(userName);
				throw new DevOpsException(DevOps.UC.USER_PASSWORD_INCORRECT);
			}
		}
	}

	@Override
	public void logout() throws Exception {
		UserInfo userInfo = DataContextThreadLocal.getUserInfo();
		operationLogger.logOther(ScopeType.PLATFORM.getDbValue(), null, "Logout: {0}", new Object[] {userInfo.getUserName()}, 
				User.QNAME, "userName", userInfo.getUserName());
	}

	@Override
	public UserResult modifyPassword(String userName, String oldPassword, String newPassword) throws Exception {
		UserResult result = new UserResult();
		result.setResultCode(0);
		if (userName.equals(UcConstants.USER_NAME_SYSADMIN)) {
			employeeService.modifyEmployeePassword(userName, oldPassword, newPassword);
			return result;
		}
		User user = baseDao.getEntity(User.QNAME, userName);
		if (user == null) {
			throw new DevOpsException(DevOps.UC.USER_NOT_EXISTED, userName);
		}
		String loginAuthenticateModel = businessDictService.getConfigParamValue(null, "Uc.LoginAuthenticateModel");
		ILdapServer ldapServer = null;
		boolean getLdapServer = false;
		if (StringUtils.isBlank(loginAuthenticateModel)) {
			ldapServer = getLdapserver(userName);
			getLdapServer = true;
			if (ldapServer == null) {
				employeeService.modifyEmployeePassword(userName, oldPassword, newPassword);
				return result;
			} else {
				ldapServer.modifyPassword(userName, oldPassword, newPassword);
				employeeService.modifyEmployeePassword(user, newPassword);
				return result;
			}
		}
		if ("local".equals(loginAuthenticateModel)) {
			employeeService.modifyEmployeePassword(userName, oldPassword, newPassword);
			return result;
		}
		if ("jira".equals(loginAuthenticateModel)) {
			jiraProjectManageToolService.changeUserPassword(userName, newPassword);
			employeeService.modifyEmployeePassword(user, newPassword);
			return result;
		}
		if (!getLdapServer) {
			ldapServer = getLdapserver(userName);
		}
		if (ldapServer == null) {
			throw new DevOpsException(DevOps.PCM.SYSCONF_NOT_CONFIG, "ldap");
		}
		ldapServer.modifyPassword(userName, oldPassword, newPassword);
		employeeService.modifyEmployeePassword(user, newPassword);
		return result;
	}

	@Override
	public String generateToken(String userName, Date dueTime) throws Exception {
		ValidateUtil.assertNotNullOrEmpty(userName, "userName");
		User user = User.FACTORY.create();
		user.setUserName(userName);
	    user = baseDao.getEntity(user);
		if (dueTime == null) {
			dueTime = new Date(System.currentTimeMillis() + businessDictService.getConfigParamValue(null, "Pcm.TokenExpiration", int.class) * 1000 * 60);
		}
		String token = null;
		Claims claims = null;
		String tenantId = null;
		if(user == null) {
			throw new DevOpsException(DevOps.UC.USER_NOT_EXISTED, userName);
		}else {
			claims = new DefaultClaims();
			claims.setSubject(userName);
			claims.put(JwtTokenService.CLAIM_KEY_EMP_ID, user.getEmployee().getEmpId());
			claims.put(JwtTokenService.CLAIM_KEY_EMP_NAME, user.getEmployee().getEmpName());
			tenantId =user.getTenantId();
			if(StringUtils.isBlank(tenantId)) {
			   tenantId = Constants.DEFAULT_TENANT_ID;//租户ID,暂用默认
			}
			claims.put(JwtTokenService.CLAIM_KEY_TENENT_ID, tenantId);
			token = jwtTokenService.generateToken(claims, dueTime);
		}
		return token;
	}

	@Override
	public void validateToken(String token) throws Exception {
		ValidateUtil.assertNotNullOrEmpty(token, "token");
		jwtTokenService.validateToken(token);
	}

	@Override
	public void unlockLogin(String userName) throws Exception {
		ValidateUtil.assertNotNullOrEmpty(userName, "userName");
		TargetRunInfo targetRunInfo = TargetRunInfo.FACTORY.create();
		targetRunInfo.setTargetType(RunTargetType.USER.getDbValue());
		targetRunInfo.setTargetId(userName);
		targetRunInfo = baseDao.getEntity(targetRunInfo);
		doUnlockLogin(targetRunInfo);
	}

	private void doUnlockLogin(TargetRunInfo targetRunInfo) throws Exception {
		if (targetRunInfo == null) {
			return;
		}
		int consecutiveLoginFailCount = CommonUtil.getExtendValue(targetRunInfo, CONSECUTIVE_LOGIN_FAIL_COUNT_KEY, -1);
		if(consecutiveLoginFailCount <= 0) {
			return;
		}
		lockService.lockAndAutoUnlock("User.login.fail", targetRunInfo.getTargetId(), -1, new Callable<Void>() {
			@Override
			public Void call() throws Exception {
				if (baseDao.getEntity(targetRunInfo) != null) {
					CommonUtil.setExtendValue(targetRunInfo, CONSECUTIVE_LOGIN_FAIL_COUNT_KEY, 0);
					CommonUtil.setExtendValue(targetRunInfo, LOGIN_FAIL_TIME_KEY, null);
					CommonUtil.setExtendValue(targetRunInfo, USER_LOCKED_FLAG_KEY, "false");
					baseDao.updateEntity(targetRunInfo);
				}
				return null;
			}
		});
	}
}
